// ZX Spectrum for Altera DE1
//
// Copyright (c) 2009-2011 Mike Stirling
//
// All rights reserved
//
// Redistribution and use in source and synthezised forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// * Redistributions of source code must retain the above copyright notice,
//   this list of conditions and the following disclaimer.
//
// * Redistributions in synthesized form must reproduce the above copyright
//   notice, this list of conditions and the following disclaimer in the
//   documentation and/or other materials provided with the distribution.
//
// * Neither the name of the author nor the names of other contributors may
//   be used to endorse or promote products derived from this software without
//   specific prior written agreement from the author.
//
// * License is granted for non-commercial use only.  A fee may not be charged
//   for redistributions as source code or in synthesized/hardware form without
//   specific prior written agreement from the author.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//

// PS/2 scancode to Spectrum matrix conversion
module keyboard
(
	input             reset,
	input             restart,
	input             clk_sys,

	input      [10:0] ps2_key,

	input      [15:0] addr,
	output      [7:0] key_data,
	input             autostart,

	output reg [11:1] Fn = 0,
	output reg  [2:0] mod = 0
);

reg  [7:0] keys[8:0];

wire [7:0] kcode       = {ps2_key[7:0]};
wire       extkey      = ps2_key[8];
wire       press_btn   = ps2_key[9];

assign key_data = ({8{addr[8]}}  | keys[0]) & ({8{addr[9]}}  | keys[1]) & ({8{addr[10]}} | keys[2]) & ({8{addr[11]}} | keys[3])
                 &({8{addr[12]}} | keys[4]) & ({8{addr[13]}} | keys[5]) & ({8{addr[14]}} | keys[6]) & ({8{addr[15]}} | keys[7])
                 &({8{~&addr[15:8]}} | keys[8]);

wire anykey = &(keys[0] & keys[1] & keys[2] & keys[3] & keys[4] & keys[5] & keys[6] & keys[7] & keys[8]);

reg  [15:0] Kp;     // Keypad 0..9, Keypad ., +, -, *, /, Enter
reg  [7:0]  others; // 0, 1 = Left/Right Host Shift; 2, 3 = Left/Right Host Ctrl; 4 = Minus, 5 = Plus, 6 = Period, 7 = Return
wire release_btn = ~press_btn;

always @(posedge clk_sys) begin
	reg auto_en;
	reg old_anykey, old_auto;
	reg old_reset = 0;
	reg old_stb;

	old_anykey <= anykey;
	old_auto <= autostart;
	if(restart) auto_en <= 1;
		else if(old_anykey & ~anykey) auto_en <=0;
	if(~old_auto & autostart & auto_en) keys[2][7] <= 0;
	if(old_auto & ~autostart) keys[2][7] <= 1;

	old_stb <= ps2_key[10];

	/*
		Host keys are stored in a variety of registers, with positive logic:

			mod[1] is the Alt key.
			Fn[x] are the Function keys.
			Kp[x] are the Keypad keys.
			others[x] are some selected non-Keypad keys

		Sam keys are stored in keys[x][y], with negative logic.
	*/
	case(kcode)
		8'h12 : others[0] <= press_btn; // left shift
		8'h59 : others[1] <= press_btn; // right shift
		8'h11 : if(~extkey) mod[1] <= press_btn; // left alt
		8'h14 :
			if(~extkey) others[2] <= press_btn; // left ctrl
			else others[3] <= press_btn; // right ctrl
		8'h05 : Fn[1]  <= press_btn; // F1
		8'h06 : Fn[2]  <= press_btn; // F2
		8'h04 : Fn[3]  <= press_btn; // F3
		8'h0C : Fn[4]  <= press_btn; // F4
		8'h03 : Fn[5]  <= press_btn; // F5
		8'h0B : Fn[6]  <= press_btn; // F6
		8'h83 : Fn[7]  <= press_btn; // F7
		8'h0A : Fn[8]  <= press_btn; // F8
		8'h01 : Fn[9]  <= press_btn; // F9
		8'h09 : Fn[10] <= press_btn; // F10
		8'h78 : Fn[11] <= press_btn; // F11
	endcase

	mod[0] <= others[0] | others[1]; // either shift
	mod[2] <= others[2] | others[3]; // either ctrl

	old_reset <= reset;
	if(~old_reset & reset)begin
		keys[0] <= 'hFF;
		keys[1] <= 'hFF;
		keys[2] <= 'hFF;
		keys[3] <= 'hFF;
		keys[4] <= 'hFF;
		keys[5] <= 'hFF;
		keys[6] <= 'hFF;
		keys[7] <= 'hFF;
		keys[8] <= 'hFF;
	end else begin
		if(old_stb != ps2_key[10]) begin

			// Modifier keys
			keys[0][0] <= ~(others[0] | others[1] | Kp[13] | Kp[14]); // Sam Shift = Host Shift or Keypad *, /
			keys[7][1] <= ~(others[2] | others[3]); // Sam Symbol = Host Ctrl
			keys[8][0] <= ~mod[1]; // Sam Ctrl = Host LAlt

			// Keypad
			keys[0][5] <= ~(Fn[1] | Kp[1]);      // Sam F1 = Host F1 | Keypad 1
			keys[0][6] <= ~(Fn[2] | Kp[2]);      // Sam F2 = ...
			keys[0][7] <= ~(Fn[3] | Kp[3]);      // Sam F3
			keys[1][5] <= ~(Fn[4] | Kp[4]);      // Sam F4
			keys[1][6] <= ~(Fn[5] | Kp[5]);      // Sam F5
			keys[1][7] <= ~(Fn[6] | Kp[6]);      // Sam F6
			keys[2][5] <= ~(Fn[7] | Kp[7]);      // Sam F7
			keys[2][6] <= ~(Fn[8] | Kp[8]);      // Sam F8 = ...
			keys[2][7] <= ~(Fn[9] | Kp[9]);      // Sam F9 = Host F9 | Keypad 9
			keys[5][7] <= ~(Fn[10]| Kp[0]);      // Sam F0 = Host F10 | Keypad 0
			keys[6][0] <= ~(others[7] | Kp[15]); // Sam Enter = Return | KP Enter

			// Plus/Minus/Dot
			keys[4][5] <= ~(others[4] | Kp[12] | Kp[14]);  // Sam Minus = Host - | Kp - | Kp /
			keys[4][6] <= ~(others[5] | Kp[11] | Kp[13]);  // Sam Plus = Host = | Kp + | Kp *
			keys[7][6] <= ~(others[6] | Kp[10]);           // Sam . = Host . | Kp .

			case(kcode)
				8'h4e : others[4] = press_btn; // Host -
				8'h55 : others[5] = press_btn; // Host =
				8'h49 : others[6] = press_btn; // Host .
			endcase

			if(extkey) begin
				// Cursor, extended KP keys
				case(kcode)
					8'h75 : keys[8][1] <= release_btn;      // Up
					8'h72 : keys[8][2] <= release_btn;      // Down
					8'h6B : keys[8][3] <= release_btn;      // Left
					8'h74 : keys[8][4] <= release_btn;      // Right

					8'h4a : Kp[14] <= press_btn;            // Keypad /
					8'h5a : Kp[15] <= press_btn;            // Keypad Enter
				endcase
			end else begin
				// Keypad, normal Return, Inv
				case(kcode)
					8'h70 : Kp[0] <= press_btn;  // Keypad 0
					8'h69 : Kp[1] <= press_btn;  // Keypad 1
					8'h72 : Kp[2] <= press_btn;  // Keypad 2
					8'h7a : Kp[3] <= press_btn;  // Keypad 3
					8'h6b : Kp[4] <= press_btn;  // Keypad 4
					8'h73 : Kp[5] <= press_btn;  // Keypad 5
					8'h74 : Kp[6] <= press_btn;  // Keypad 6
					8'h6c : Kp[7] <= press_btn;  // Keypad 7
					8'h75 : Kp[8] <= press_btn;  // Keypad 8
					8'h7d : Kp[9] <= press_btn;  // Keypad 9
					8'h71 : Kp[10] <= press_btn; // Keypad .
					8'h7b : Kp[11] <= press_btn; // Keypad -
					8'h7d : Kp[12] <= press_btn; // Keypad +
					8'h7c : Kp[13] <= press_btn; // Keypad *

					8'h5a : others[7] <= press_btn; // Return
					8'h4a : keys[7][7] <= release_btn; // Sam INV = Host /
				endcase
			end

			case(kcode)
				// Letters and others
				8'h1a : keys[0][1] <= release_btn; // Z
				8'h22 : keys[0][2] <= release_btn; // X
				8'h21 : keys[0][3] <= release_btn; // C
				8'h2a : keys[0][4] <= release_btn; // V

				8'h1c : keys[1][0] <= release_btn; // A
				8'h1b : keys[1][1] <= release_btn; // S
				8'h23 : keys[1][2] <= release_btn; // D
				8'h2b : keys[1][3] <= release_btn; // F
				8'h34 : keys[1][4] <= release_btn; // G

				8'h15 : keys[2][0] <= release_btn; // Q
				8'h1d : keys[2][1] <= release_btn; // W
				8'h24 : keys[2][2] <= release_btn; // E
				8'h2d : keys[2][3] <= release_btn; // R
				8'h2c : keys[2][4] <= release_btn; // T

				8'h16 : keys[3][0] <= release_btn; // 1
				8'h1e : keys[3][1] <= release_btn; // 2
				8'h26 : keys[3][2] <= release_btn; // 3
				8'h25 : keys[3][3] <= release_btn; // 4
				8'h2e : keys[3][4] <= release_btn; // 5

				8'h45 : keys[4][0] <= release_btn; // 0
				8'h46 : keys[4][1] <= release_btn; // 9
				8'h3e : keys[4][2] <= release_btn; // 8
				8'h3d : keys[4][3] <= release_btn; // 7
				8'h36 : keys[4][4] <= release_btn; // 6

				8'h4d : keys[5][0] <= release_btn; // P
				8'h44 : keys[5][1] <= release_btn; // O
				8'h43 : keys[5][2] <= release_btn; // I
				8'h3c : keys[5][3] <= release_btn; // U
				8'h35 : keys[5][4] <= release_btn; // Y

				8'h4b : keys[6][1] <= release_btn; // L
				8'h42 : keys[6][2] <= release_btn; // K
				8'h3b : keys[6][3] <= release_btn; // J
				8'h33 : keys[6][4] <= release_btn; // H

				8'h29 : keys[7][0] <= release_btn; // SPACE
				8'h3a : keys[7][2] <= release_btn; // M
				8'h31 : keys[7][3] <= release_btn; // N
				8'h32 : keys[7][4] <= release_btn; // B

				8'h76 : keys[3][5] <= release_btn; // Escape
				8'h0D : keys[3][6] <= release_btn; // TAB
				8'h58 : keys[3][7] <= release_btn; // Caps lock
				8'h66 : keys[4][7] <= release_btn; // Backspace (DEL)

				8'h54 : keys[5][5] <= release_btn; // Sam = = Host [
				8'h5b : keys[5][6] <= release_btn; // Sam " = Host ]
				8'h4c : keys[6][5] <= release_btn; // Sam ; = Host ;
				8'h52 : keys[6][6] <= release_btn; // Sam : = Host '
				8'h41 : keys[7][5] <= release_btn; // Sam , = Host ,

				8'h11 : if(extkey) keys[6][7] = release_btn;
			endcase
		end
	end
end
endmodule
